import locale

from django.contrib import admin, messages
from django.core import serializers
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from ...tools.utils import locale_currency
from ..db.models import OwnerModel


class PreviewMixin:
    image_width = 100
    image_height = 100

    @admin.display(description=_('preview'))
    def get_preview(self, obj):
        html = '<br>'.join('''<div style="width: {1}; height: {2}; overflow: hidden;">
                 <a href="{0}" rel="external" target="_blank">
                   <img src="{0}" max-width="{1}" max-height="{2} width="{1}" height="{2}">
                 </a>
               </div>'''.format(item.strip(), self.image_width, self.image_height) for item in obj.images.split('\n'))
        return format_html(html)


class CurrencyMixin:

    def format_currency(self, amount, min_value, currency='CNY'):
        value = amount / 100
        formatted = locale_currency(currency, value)
        if value < min_value:
            formatted = f'{formatted}<span style="color: red;"> (低于{locale.currency(min_value, False, True)})</span>'
        return format_html(formatted)


class SuperAdmin(CurrencyMixin, PreviewMixin, admin.ModelAdmin):
    readonly_fields = []  # 默认从tuple改为list

    def get_readonly_fields(self, request, obj=None):
        if obj:
            return [f.name for f in obj._meta.get_fields() if not f.editable] + self.readonly_fields
        return super().get_readonly_fields(request)

    def get_sortable_by(self, request):
        if self.readonly_fields and isinstance(self.readonly_fields, tuple):
            messages.warning(request, 'readonly_fields 为 tuple, 建议改为 list, 修改页面未容错，会报错')
        return super().get_sortable_by(request)

    def save_model(self, request, obj, form, change):
        if not obj.pk:  # 如果obj.pk不存在, 为新创建
            if isinstance(obj, OwnerModel):
                obj.username = request.user.username
                obj.user_avatar = request.user.avatar
        super().save_model(request, obj, form, change)

    def log_change(self, request, obj, message):
        new_message = f'json: {serializers.serialize("json", [obj])}'
        super().log_change(request, obj, new_message)



import datetime
import hashlib
import hmac
import mimetypes
import uuid
from base64 import b64encode
from urllib.parse import urlparse

from django.conf import settings
from rest_framework.exceptions import APIException, ValidationError
from rest_framework.permissions import AllowAny, IsAuthenticated
from rest_framework.response import Response
from rest_framework.utils import json
from rest_framework.views import APIView

from ..django.db.utils import timestamp
from .serializers import DownloadUrlSerializer, UploadParamsSerializer


def create_filename(filename):
    ext = filename.split('.')[-1]
    return f'{uuid.uuid4().hex}.{ext}'


def sign(key, msg):
    return hmac.new(key, msg.encode('utf-8'), hashlib.sha256).digest()


def get_signing_key(key, date_stamp, region, service):
    kDate = sign(('AWS4' + key).encode('utf-8'), date_stamp)
    kRegion = sign(kDate, region)
    kService = sign(kRegion, service)
    kSigning = sign(kService, 'aws4_request')
    return kSigning


def get_signature(msg, digestmod):
    if not settings.CLOUD_STORAGE_SECRET:
        raise APIException('CLOUD_STORAGE_SECRET 不存在')
    key = settings.CLOUD_STORAGE_SECRET.encode()
    return b64encode(hmac.new(key, msg.encode('utf-8'), digestmod).digest()).decode()


def get_params(cloud, region, bucket, filename, rename, expiration, content_encoding, cache_control):
    content_type = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
    path = f'upload/{create_filename(filename)}' if rename else filename  # 是否重命名
    # 计算policy
    conditions = [{
        'bucket': bucket
    }, ['starts-with', '$key', path.split('/')[0]], ['starts-with', '$Content-Type', content_type]]
    policy_dict = {
        'expiration':
            (datetime.datetime.utcnow() + datetime.timedelta(hours=expiration)).strftime('%Y-%m-%dT%H:%M:%S.000Z'),
        'conditions':
            conditions
    }
    # 时间相关
    t = datetime.datetime.utcnow()
    expire = t + datetime.timedelta(days=1)
    amz_date = t.strftime('%Y%m%dT%H%M%SZ')
    date_stamp = t.strftime('%Y%m%d')
    if cloud == 'aws':
        conditions += [{
            'acl': 'public-read'
        }, {
            'success_action_status': '204'
        }, {
            'x-amz-meta-uuid': '14365123651274'
        }, {
            'x-amz-server-side-encryption': 'AES256'
        }, ['starts-with', '$x-amz-meta-tag', ''], {
            'x-amz-credential': f'{settings.CLOUD_STORAGE_ID}/{date_stamp}/{region}/s3/aws4_request'
        }, {
            'x-amz-algorithm': 'AWS4-HMAC-SHA256'
        }, {
            'x-amz-date': amz_date
        }]
        policy_dict['conditions'] = conditions
        # policy_dict['conditions'].append(['content-length-range', 1, 1024 * 1024 * 4])
    if content_encoding == 'gzip':
        policy_dict['conditions'].append(['starts-with', '$Content-Encoding', content_encoding])
    string_to_sign = b64encode(json.dumps(policy_dict).encode('utf-8'))
    #
    params = {'key': path, 'Content-Type': content_type, 'policy': string_to_sign}
    match cloud:
        case 'aliyun':  # 阿里云
            signature = b64encode(
                hmac.new(settings.CLOUD_STORAGE_SECRET.encode('utf-8'), string_to_sign, hashlib.sha1).digest())
            params.update({'OSSAccessKeyId': settings.CLOUD_STORAGE_ID, 'signature': signature})
        case 'aws':  # AWS https://docs.aws.amazon.com/AmazonS3/latest/API/sigv4-HTTPPOSTConstructPolicy.html
            for condition in conditions:
                if isinstance(condition, dict):
                    params.update(condition)
            params.pop('bucket')
            signing_key = get_signing_key(settings.CLOUD_STORAGE_SECRET, date_stamp, region, 's3')
            signature = hmac.new(signing_key, string_to_sign, hashlib.sha256).hexdigest()  # 16进制
            params.update({'x-amz-meta-tag': '', 'x-amz-signature': signature})
    # 其他 Content-Encoding, Cache-Control
    if content_encoding == 'gzip':
        params['Content-Encoding'] = content_encoding
    if cache_control:
        params['Cache-Control'] = cache_control
    return params


class DownloadUrlView(APIView):
    """
    获取下载地址
    """
    lookup_field = 'cloud'
    serializer_class = DownloadUrlSerializer

    def get(self, request, *args, **kwargs):
        url = request.GET.get(
            'url')  # https://documents-domain-com.oss-cn-shanghai.aliyuncs.com/contract/001/Linux_Command3.pdf
        if not url:
            raise ValidationError('url不能为空')
        url_components = urlparse(url)
        bucket = url_components.netloc.replace(f'.{settings.CLOUD_STORAGE_DOMAIN_NAME}', '')
        expires = int(timestamp(datetime.datetime.now())) + 600  # 10分钟有效
        string_to_sign = f'GET\n\n\n{expires}\n/{bucket}{url_components.path}'  # 字符串
        signature = get_signature(string_to_sign, hashlib.sha1)
        params = {'url': f'{url}?OSSAccessKeyId={settings.CLOUD_STORAGE_ID}&Expires={expires}&Signature={signature}'}
        return Response(params)


class UploadParamsView(APIView):
    """
    获取上传参数
    filename -- 文件名, 可含有路径
    expiration -- (可选)过期时间, 默认50年
    content_encoding -- (可选)默认不压缩
    cache_control -- (可选)默认没有
    """
    lookup_field = 'cloud'
    serializer_class = UploadParamsSerializer  # 加了只是为OpenAPI用, 不影响输出结果

    def post(self, request, *args, **kwargs):
        filename = request.data.get('filename')
        if not filename:
            raise ValidationError('filename不能为空')
        bucket = request.data.get('bucket', settings.BUCKET_MEDIA)
        expiration = int(request.data.get('expiration', 24 * 365 * 50))  # 过期时间
        content_encoding = request.data.get('content_encoding', '')
        cache_control = request.data.get('cache_control')
        rename = False
        match bucket:
            case settings.BUCKET_MEDIA:
                endpoint = settings.MEDIA_URL
                rename = True
            case settings.BUCKET_STATIC:  # 传到static下的不修改大小写
                endpoint = settings.STATIC_URL
                # filename = filename.lower()
            case settings.BUCKET_CLOUD:
                endpoint = settings.CLOUD_URL
            case _:
                endpoint = ''
        params = get_params(kwargs.get(self.lookup_field), settings.CLOUD_STORAGE_REGION, bucket, filename, rename,
                            expiration, content_encoding, cache_control)
        return Response({'endpoint': endpoint, 'params': params})



import os
from urllib.parse import unquote_plus

from django.conf import settings
from django.forms import widgets
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.safestring import mark_safe


class CloudFileWidget(widgets.TextInput):

    class Media:
        js = ['s3direct/dist/index.js']
        css = {'all': ['s3direct/dist/index.css']}

    def __init__(self, *args, **kwargs):
        self.dest = kwargs.pop('dest', None)
        super().__init__(*args, **kwargs)

    def render(self, name, value, **kwargs):
        file_url = value or ''
        csrf_cookie_name = getattr(settings, 'CSRF_COOKIE_NAME', 'csrftoken')
        ctx = {
            'policy_url': reverse('upload-params', args=['aliyun']),
            'signing_url': reverse('s3direct-signing'),
            'dest': self.dest,
            'name': name,
            'csrf_cookie_name': csrf_cookie_name,
            'file_url': file_url,
            'file_name': os.path.basename(unquote_plus(file_url))
        }
        return mark_safe(render_to_string(os.path.join('s3direct', 's3direct-widget.tpl'), ctx))



import datetime

from django.apps.registry import apps
from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.urls import include, path, re_path
from django.utils.translation import gettext_lazy as _
from django.views.decorators.cache import cache_page
from django.views.generic.base import RedirectView
from django.views.i18n import JavaScriptCatalog
from rest_framework_simplejwt.views import token_obtain_pair, token_refresh

from .rest_framework.views import DownloadUrlView, UploadParamsView

# Django
urlpatterns = [
    path('admin/', admin.site.urls),
    path('i18n/', include('django.conf.urls.i18n')),
    path('jsi18n/',
         cache_page(86400,
                    key_prefix=f'js18n-{datetime.datetime.now().strftime("%Y%m%d")}')(JavaScriptCatalog.as_view()),
         name='javascript-catalog')
]
# Vendor
urlpatterns += [
    path('api-auth/', include('rest_framework.urls')),
    path('api/token/', token_obtain_pair, name='token_obtain_pair'),
    path('api/token/refresh/', token_refresh, name='token_refresh'),
    # path('search/', include('haystack.urls'))
]
# Railgun S
urlpatterns += [
    re_path(r'^download-url/(?P<cloud>(aliyun|aws))/$', DownloadUrlView.as_view(), name='download-url'),
    re_path(r'^upload-params/(?P<cloud>(aliyun|aws))/$', UploadParamsView.as_view(), name='upload-params'),
    path('favicon.ico', RedirectView.as_view(url=f'{settings.STATIC_URL}favicon.ico', permanent=True)),
]

urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)

admin.site.site_header = _('app_name')
admin.site.site_title = _('app_name')



import datetime
import time
from typing import Optional

from django.conf import settings
from django.utils import timezone


def timestamp(date):
    return time.mktime(date.timetuple())


def generate_shard_id(user_id: int) -> int:
    time_offset = int(timestamp(datetime.datetime.now()) - 1510358400)  # 2017-11-11 00:00:00 中国时间
    return user_id | time_offset << 32


def get_object(model, using='default', default=None, **kwargs):
    try:
        return model.objects.using(using).get(**kwargs)
    except model.DoesNotExist:
        return default


def get_user_id(pk: int) -> int:
    """
    需要保证传入的pk都是int
    """
    return pk & 0xFFFFFFFF


def db_master(user_id: Optional[int] = None) -> str:
    """
    需要保证传入的user_id都是int
    """
    if not user_id:
        return 'default'
    else:
        if user_id < 100001:
            return 'default'
        else:
            return f'db_{user_id % settings.SHARD_COUNT}'


def db_slave(user_id: Optional[int] = None) -> str:
    suffix = ''
    return f'{db_master(user_id)}{suffix}'  # replica_


def redis_master(user_id: Optional[int] = None) -> int:
    if not user_id:
        return 0
    else:
        if user_id < 100001:
            return 0
        else:
            return int(user_id / 1000)


def datetime_to_unixtime(date_time) -> int:
    date_string = date_time.strftime('%Y-%m-%d %H:%M:%S')
    date_time = datetime.datetime.strptime(date_string, '%Y-%m-%d %H:%M:%S') + datetime.timedelta(hours=8)
    return int(time.mktime(date_time.timetuple()))


def unixtime_to_datetime(local_time):
    date_string = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(local_time))
    date_time = datetime.datetime.strptime(date_string, '%Y-%m-%d %H:%M:%S')
    return date_time


def unixtime_to_date(local_time):
    return time.strftime('%Y-%m-%d', time.localtime(local_time))


def datetime_timezone_zero():
    return datetime.datetime(timezone.now().year, timezone.now().month, timezone.now().day, 0, 0, 0)


def string_to_unixtime(string) -> int:
    date_time = datetime.datetime.strptime(string, '%Y-%m-%d %H:%M:%S')
    return int(time.mktime(date_time.timetuple()))



from typing import Optional
from urllib.parse import urlparse

from django.conf import settings
from django.http import HttpRequest
from django.shortcuts import render
from django.utils.translation import gettext, gettext_lazy as _
from django.views.generic.base import TemplateView


def generate_link(urlstring: str, request: Optional[HttpRequest]) -> str:
    if not urlstring or urlparse(urlstring).scheme:  # 如果为空或有scheme
        return urlstring
    else:
        if request:
            app_scheme = request.headers.get('App-Scheme')
            if app_scheme:
                return f'{app_scheme}://{urlstring}'
        return f'/{urlstring}/' if urlstring != '/' else '/'


class WebView(TemplateView):
    name: Optional[str] = None
    title: Optional[str] = None
    endpoint: Optional[str] = ''

    def get(self, request, *args, **kwargs):
        if self.name.endswith('_create'):  # 创建页
            verbose_name = ' '.join(gettext(item) for item in self.name.split('_')[::-1])  # NY: 用gettext_lazy会报错
        else:
            verbose_name = _(self.name.replace('y_list', 'ies').replace('_list', 's').replace('_detail', ''))
        title = self.title if self.title is not None else verbose_name  # 用 is not None 才能传入空标题
        if self.endpoint is not None:
            version = settings.REST_FRAMEWORK.get('DEFAULT_VERSION', '')
            endpoint = self.endpoint or f'/api/{version}{request.get_full_path()}'
            dataset_endpoint = f' data-endpoint="{endpoint}"' if endpoint else ''
        template_name = self.template_name or f'web/{self.name}.html'
        extras = kwargs
        is_weixin = 'MicroMessenger' in request.headers['User-Agent']
        return render(request, template_name, locals())



from setuptools import find_packages, setup


with open('README.md', encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='RailgunS',
    version='0.57',
    author='NY',
    author_email='nyssance@icloud.com',
    description='Only My Railgun',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/nyssance/railguns',
    project_urls={
        'Bug Tracker': 'https://github.com/nyssance/railguns/issues'
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django :: 4.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Internet :: WWW/HTTP'
    ],
    # package_dir={'': 'src'},
    packages=find_packages(exclude=['tests']),
    python_requires='>=3.10',
    # zip_safe=False,
    include_package_data=True,
    install_requires=[
        'django[argon2]==4.0.5',
        'djangorestframework==3.13.1',
        'djangorestframework_simplejwt==5.2.0',
        'gunicorn==20.1.0',
        'uvicorn==0.17.6',
        #
        'django-ckeditor==6.4.0',
        'django-filter==21.1',
        'django-htmlmin==0.11.0',
        'hiredis==2.0.0',
        'mysqlclient==2.1.0',
        'redis==4.3.3'
    ],
    extras_require={
        'dev': [],
        'prod': []
    }
)



from django.utils.html import format_html


# django 自带的 format_html_join 方法非常鸡肋
def format_html_field_join(field, format_string='{}', join_string='<br>'):
    return format_html(join_string.join(format_string.format(item) for item in field.all()))


def format_html_string_join(string, format_string='{}', join_string='<br>'):
    return format_html(join_string.join(format_string.format(item) for item in filter(None, string.split(','))))
